
/*
 * FLAC (Free Lossless Audio Codec) decoder
 * Copyright (c) 2003 Alex Beregszaszi
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file libavcodec/flacdec.c
 * FLAC (Free Lossless Audio Codec) decoder
 * @author Alex Beregszaszi
 *
 * For more information on the FLAC format, visit:
 *  http://flac.sourceforge.net/
 *
 * This decoder can be used in 1 of 2 ways: Either raw FLAC data can be fed
 * through, starting from the initial 'fLaC' signature; or by passing the
 * 34-byte streaminfo structure through avctx->extradata[_size] followed
 * by data starting with the 0xFFF8 marker.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pspuser.h>

#include <limits.h>

#include "strtool.h"

#include "LibSnd_Const_Global.h"
#include "LibSnd_Const_Internal.h"

#include "ffmpeg_flac_config.h"
#include "_flac_bitstream.h"
#include "golomb.h"
#include "flac.h"

#undef NDEBUG
#include <assert.h>

static const int sample_rate_table[] =
{ 0,
  88200, 176400, 192000,
  8000, 16000, 22050, 24000, 32000, 44100, 48000, 96000,
  0, 0, 0, 0 };

static const int sample_size_table[] =
{ 0, 8, 12, 0, 16, 20, 24, 0 };

static const int blocksize_table[] = {
     0,    192, 576<<0, 576<<1, 576<<2, 576<<3,      0,      0,
256<<0, 256<<1, 256<<2, 256<<3, 256<<4, 256<<5, 256<<6, 256<<7
};

// -------------------------------------------------------------------------------

static struct FLACContext s;
static FILE *FileHandle;

// -------------------------------------------------------------------------------

static int64_t get_utf8(GetBitContext *gb)
{
    s32 val=get_bits(gb, 8);
    int ones= 7 - ff_log2_tab[val ^ 255];
    if(ones==1) return -1;
    
    val&= 127>>ones;
    
    while(--ones > 0){
        int tmp= get_bits(gb, 8) - 128;
        if(tmp>>6) return -1;
        val= (val<<6) + tmp;
    }
    
    return val;
}

// -------------------------------------------------------------------------------

TVorbisComment VorbisComment;

static void VorbisComment_Init(void)
{
  VorbisComment.LinesCount=0;
  for(u32 idx=0;idx<VorbisComment_LinesMax;idx++){
    VorbisComment.pLines[idx]=NULL;
  }
}

static void VorbisComment_Free(void)
{
  VorbisComment.LinesCount=0;
  for(u32 idx=0;idx<VorbisComment_LinesMax;idx++){
    if(VorbisComment.pLines[idx]!=NULL){
      free(VorbisComment.pLines[idx]); VorbisComment.pLines[idx]=NULL;
    }
  }
}

static void VorbisComment_Parse(char *pbuf,u32 bufsize)
{
  u32 bufidx=0;

#define Get32bit(ret) { \
  u8 *p=(u8*)pbuf; \
  ret=(*p++)<<0; \
  ret|=(*p++)<<8; \
  ret|=(*p++)<<16; \
  ret|=(*p++)<<24; \
  pbuf+=4; \
  printf("ret=$%x\n",ret);\
}
  
  {
    char encname[256];
    u32 len;
    Get32bit(len);
    for(u32 idx=0;idx<len;idx++){
      if((idx+1)==256) break;
      encname[idx]=pbuf[idx];
      encname[idx+1]=0;
    }
    printf("Encoder name: '%s'\n",encname);
    pbuf+=len;
  }
  
  u32 ItemsCount;
  Get32bit(ItemsCount);
  
  for(u32 idx=0;idx<ItemsCount;idx++){
    char line[256];
    bool HaveEqual=false;
    char *pbodyline=NULL;
    u32 len;
    Get32bit(len);
    for(u32 idx=0;idx<len;idx++){
      if((idx+1)==256) break;
      char ch=pbuf[idx];
      if((ch=='=')&&(HaveEqual==false)){
        HaveEqual=true;
        pbodyline=&line[idx+1];
        ch=0;
      }
      line[idx]=ch;
      line[idx+1]=0;
    }
    if((HaveEqual==true)&&(str_isEmpty(pbodyline)==false)){
      printf("ItemName='%s', Body='%s'.\n",line,pbodyline);
      bool addf=false;
      if(isStrEqual_NoCaseSensitive(line,"album")==true) addf=true;
      if(isStrEqual_NoCaseSensitive(line,"artist")==true) addf=true;
      if(isStrEqual_NoCaseSensitive(line,"comment")==true) addf=true;
      if(isStrEqual_NoCaseSensitive(line,"title")==true) addf=true;
      if(addf==true) VorbisComment.pLines[VorbisComment.LinesCount++]=str_AllocateCopy(pbodyline);
    }
    pbuf+=len;
  }
}

// -------------------------------------------------------------------------------

static void parse_streaminfo(struct FLACContext *s)
{
  u8 buffer[FLAC_STREAMINFO_SIZE];
  fread(buffer,1,FLAC_STREAMINFO_SIZE,FileHandle);
  
  GetBitContext gb;
  init_get_bits(&gb, buffer, FLAC_STREAMINFO_SIZE*8);

  /* mandatory streaminfo */
  s->min_blocksize = get_bits(&gb, 16);
  s->max_blocksize = get_bits(&gb, 16);

  skip_bits(&gb, 24); /* skip min frame size */
  s->max_framesize = get_bits_long(&gb, 24);

  s->samplerate = get_bits_long(&gb, 20);
  s->channels = get_bits(&gb, 3) + 1;
  s->bps = get_bits(&gb, 5) + 1;

  s->samples  = get_bits_long(&gb, 32) << 4;
  s->samples |= get_bits_long(&gb, 4);

  skip_bits(&gb, 64); /* md5 sum */
  skip_bits(&gb, 64); /* md5 sum */
}

static void dump_streaminfo(struct FLACContext *s)
{
  printf("Blocksize: %d .. %d\n", s->min_blocksize, s->max_blocksize);
  printf("Max Framesize: %d\n", s->max_framesize);
  printf("Samplerate: %d\n", s->samplerate);
  printf("Channels: %d\n", s->channels);
  printf("Bits: %d\n", s->bps);
  printf("SamplesCount/256: %d\n", (u32)(s->samples/256));
}

static bool parse_metadata(struct FLACContext *s)
{
    bool streaminfo_updated=false;
    
    while(1){
        u8 data[4];
        fread(data,1,4,FileHandle);
        u32 last = data[0]&0x80;
        u32 type = data[0]&0x7f;
        u32 size = (data[1]<<16)|(data[2]<<8)|(data[3]<<0);

        if (size) {
          switch (type) {
            case FLAC_METADATA_TYPE_STREAMINFO: {
              parse_streaminfo(s);
              streaminfo_updated = true;
            } break;
            case FLAC_METADATA_TYPE_VORBIS_COMMENT: {
              u8 *pbuf=(u8*)malloc(size);
              fread(pbuf,1,size,FileHandle);
              VorbisComment_Parse((char*)pbuf,size);
              if(pbuf!=NULL){
                free(pbuf); pbuf=NULL;
              }
            } break;
            default: {
              printf("Skip metadata: Last=%d, Type=$%02x, Size=%d.\n",last,type,size);
              fseek(FileHandle,ftell(FileHandle)+size,SEEK_SET);
            } break;
          }
        }
        if(last!=0) break;
    }

    if (streaminfo_updated==false){
      printf("Can not found streaminfo.\n");
      return(false);
    }
    
    return(true);
}

struct FLACContext* flac_open(FILE *fh)
{
  MemSet8CPU(0,&s,sizeof(struct FLACContext));
  
  VorbisComment_Init();
  
  FileHandle=fh;
  
  {
    u8 ID[4];
    fread(ID,1,4,FileHandle);
    if((ID[0]=='f')&&(ID[1]=='L')&&(ID[2]=='a')&&(ID[3]=='C')){
      if(parse_metadata(&s)==false) return(NULL);
      }else{
      fseek(FileHandle,0,SEEK_SET);
      parse_streaminfo(&s);
    }
  }
  
  dump_streaminfo(&s);
  
  if(s.max_blocksize==0){
    printf("Flac: Invalid max_blocksize. (%dbyte)\n",s.max_blocksize);
    return(NULL);
  }
  
  if(s.bps!=16){
    printf("Flac: Invalid bps. (%dbps)\n",s.bps);
    return(NULL);
  }
  
  if (s.max_framesize == 0 && s.max_blocksize) {
      // FIXME header overhead
      s.max_framesize= (s.channels * s.bps * s.max_blocksize + 7)/ 8;
  }
  
  for (u32 i = 0; i < s.channels; i++) {
    s.decoded[i] = (s32*)malloc(sizeof(int32_t)*s.max_blocksize);
  }
    
  return(&s);
}

void flac_close(void)
{
  for (u32 i = 0; i < s.channels; i++) {
    if(s.decoded[i]!=NULL){
      free(s.decoded[i]); s.decoded[i]=NULL;
    }
  }
  
  VorbisComment_Free();
}

// -----------------------------------------------------------------------------------------------------

static int decode_residuals(FLACContext *s, GetBitContext *gb, int channel, int pred_order)
{
    int i, tmp, partition, method_type, rice_order;
    int sample = 0, samples;

    method_type = get_bits(gb, 2);
    if (method_type > 1) {
        printf("FlacError: illegal residual coding method %d\n", method_type);
        return -1;
    }

    rice_order = get_bits(gb, 4);

    samples= s->blocksize >> rice_order;
    if (pred_order > samples) {
        printf("FlacError: invalid predictor order: %i > %i\n", pred_order, samples);
        return -1;
    }

    sample=
    i= pred_order;
    for (partition = 0; partition < (1 << rice_order); partition++) {
        tmp = get_bits(gb, method_type == 0 ? 4 : 5);
        if (tmp == (method_type == 0 ? 15 : 31)) {
            tmp = get_bits(gb, 5);
            for (; i < samples; i++, sample++)
                s->decoded[channel][sample] = get_sbits(gb, tmp);
        } else {
            for (; i < samples; i++, sample++) {
                s->decoded[channel][sample] = get_sr_golomb_flac(gb, tmp, INT_MAX, 0);
            }
        }
        i= 0;
    }

    return 0;
}

static int decode_subframe_fixed(FLACContext *s, GetBitContext *gb, int channel, int pred_order)
{
    const int blocksize = s->blocksize;
    int32_t *decoded = s->decoded[channel];
    int a=0,b=0,c=0,d=0;

    /* warm up samples */
    for (u32 i = 0; i < pred_order; i++) {
        decoded[i] = get_sbits(gb, s->curr_bps);
    }

    if (decode_residuals(s, gb, channel, pred_order) < 0)
        return -1;

    if (pred_order > 0)
        a = decoded[pred_order-1];
    if (pred_order > 1)
        b = a - decoded[pred_order-2];
    if (pred_order > 2)
        c = b - decoded[pred_order-2] + decoded[pred_order-3];
    if (pred_order > 3)
        d = c - decoded[pred_order-2] + 2*decoded[pred_order-3] - decoded[pred_order-4];

    switch (pred_order) {
    case 0:
        break;
    case 1:
        for (u32 i = pred_order; i < blocksize; i++)
            decoded[i] = a += decoded[i];
        break;
    case 2:
        for (u32 i = pred_order; i < blocksize; i++)
            decoded[i] = a += b += decoded[i];
        break;
    case 3:
        for (u32 i = pred_order; i < blocksize; i++)
            decoded[i] = a += b += c += decoded[i];
        break;
    case 4:
        for (u32 i = pred_order; i < blocksize; i++)
            decoded[i] = a += b += c += d += decoded[i];
        break;
    default:
        printf("FlacError: illegal pred order %d\n", pred_order);
        return -1;
    }

    return 0;
}

static int decode_subframe_lpc(FLACContext *s, GetBitContext *gb, int channel, int pred_order)
{
    int i, j;
    int coeff_prec, qlevel;
    int coeffs[pred_order];
    int32_t *decoded = s->decoded[channel];

    /* warm up samples */
    for (i = 0; i < pred_order; i++) {
        decoded[i] = get_sbits(gb, s->curr_bps);
    }

    coeff_prec = get_bits(gb, 4) + 1;
    if (coeff_prec == 16) {
        printf("FlacError: invalid coeff precision\n");
        return -1;
    }
    qlevel = get_sbits(gb, 5);
    if (qlevel < 0) {
        printf("FlacError: qlevel %d not supported, maybe buggy stream\n", qlevel);
        return -1;
    }

    for (i = 0; i < pred_order; i++) {
        coeffs[i] = get_sbits(gb, coeff_prec);
    }

    if (decode_residuals(s, gb, channel, pred_order) < 0)
        return -1;

    if (s->bps > 16) {
        int64_t sum;
        for (i = pred_order; i < s->blocksize; i++) {
            sum = 0;
            for (j = 0; j < pred_order; j++)
                sum += (int64_t)coeffs[j] * decoded[i-j-1];
            decoded[i] += sum >> qlevel;
        }
    } else {
        for (i = pred_order; i < s->blocksize-1; i += 2) {
            int c;
            int d = decoded[i-pred_order];
            int s0 = 0, s1 = 0;
            for (j = pred_order-1; j > 0; j--) {
                c = coeffs[j];
                s0 += c*d;
                d = decoded[i-j];
                s1 += c*d;
            }
            c = coeffs[0];
            s0 += c*d;
            d = decoded[i] += s0 >> qlevel;
            s1 += c*d;
            decoded[i+1] += s1 >> qlevel;
        }
        if (i < s->blocksize) {
            int sum = 0;
            for (j = 0; j < pred_order; j++)
                sum += coeffs[j] * decoded[i-j-1];
            decoded[i] += sum >> qlevel;
        }
    }

    return 0;
}

static inline int decode_subframe(FLACContext *s, GetBitContext *gb, int channel)
{
    int type, wasted = 0;
    int i, tmp;

    s->curr_bps = s->bps;
    if (channel == 0) {
        if (s->decorrelation == RIGHT_SIDE)
            s->curr_bps++;
    } else {
        if (s->decorrelation == LEFT_SIDE || s->decorrelation == MID_SIDE)
            s->curr_bps++;
    }

    if (get_bits1(gb)) {
        printf("FlacError: invalid subframe padding\n");
        return -1;
    }
    type = get_bits(gb, 6);

    if (get_bits1(gb)) {
        wasted = 1;
        while (!get_bits1(gb))
            wasted++;
        s->curr_bps -= wasted;
    }

//FIXME use printf2 for types
    if (type == 0) {
        tmp = get_sbits(gb, s->curr_bps);
        for (i = 0; i < s->blocksize; i++)
            s->decoded[channel][i] = tmp;
    } else if (type == 1) {
        for (i = 0; i < s->blocksize; i++)
            s->decoded[channel][i] = get_sbits(gb, s->curr_bps);
    } else if ((type >= 8) && (type <= 12)) {
        if (decode_subframe_fixed(s, gb, channel, type & ~0x8) < 0)
            return -1;
    } else if (type >= 32) {
        if (decode_subframe_lpc(s, gb, channel, (type & ~0x20)+1) < 0)
            return -1;
    } else {
        printf("FlacError: invalid coding type\n");
        return -1;
    }

    if (wasted) {
        int i;
        for (i = 0; i < s->blocksize; i++)
            s->decoded[channel][i] <<= wasted;
    }

    return 0;
}

static int decode_frame(FLACContext *s, GetBitContext *gb)
{
    int blocksize_code, sample_rate_code, sample_size_code, assignment; // , crc8;
    decorrelation_type decorrelation;
    int bps, blocksize, samplerate;

    blocksize_code = get_bits(gb, 4);

    sample_rate_code = get_bits(gb, 4);

    assignment = get_bits(gb, 4); /* channel assignment */
    if (assignment < 8 && s->channels == assignment+1)
        decorrelation = INDEPENDENT;
    else if (assignment >=8 && assignment < 11 && s->channels == 2)
        decorrelation = (decorrelation_type)(LEFT_SIDE + assignment - 8);
    else {
        printf("FlacError: unsupported channel assignment %d (channels=%d)\n", assignment, s->channels);
        return -1;
    }

    sample_size_code = get_bits(gb, 3);
    if (sample_size_code == 0)
        bps= s->bps;
    else if ((sample_size_code != 3) && (sample_size_code != 7))
        bps = sample_size_table[sample_size_code];
    else {
        printf("FlacError: invalid sample size code (%d)\n", sample_size_code);
        return -1;
    }
    if (bps > 16) {
      printf("libflac: Internal error. bps=%d.\n",bps);
      while(1);
    } else {
        s->sample_shift = 16 - bps;
        s->is32 = 0;
    }
    s->bps = bps;

    if (get_bits1(gb)) {
        printf("FlacError: broken stream, invalid padding\n");
        return -1;
    }

    if (get_utf8(gb) < 0) {
        printf("FlacError: utf8 fscked\n");
        return -1;
    }

    if (blocksize_code == 0)
        blocksize = s->min_blocksize;
    else if (blocksize_code == 6)
        blocksize = get_bits(gb, 8)+1;
    else if (blocksize_code == 7)
        blocksize = get_bits(gb, 16)+1;
    else
        blocksize = blocksize_table[blocksize_code];

    if (blocksize > s->max_blocksize) {
        printf("FlacError: blocksize %d > %d\n", blocksize, s->max_blocksize);
        return -1;
    }

    if (sample_rate_code == 0)
        samplerate= s->samplerate;
    else if (sample_rate_code < 12)
        samplerate = sample_rate_table[sample_rate_code];
    else if (sample_rate_code == 12)
        samplerate = get_bits(gb, 8) * 1000;
    else if (sample_rate_code == 13)
        samplerate = get_bits(gb, 16);
    else if (sample_rate_code == 14)
        samplerate = get_bits(gb, 16) * 10;
    else {
        printf("FlacError: illegal sample rate code %d\n", sample_rate_code);
        return -1;
    }

    skip_bits(gb, 8);
/*
    crc8 = av_crc(av_crc_get_table(AV_CRC_8_ATM), 0,
                  gb.buffer, get_bits_count(gb)/8);
    if (crc8) {
        printf("FlacError: header crc mismatch crc=%2X\n", crc8);
        return -1;
    }
*/

    s->blocksize    = blocksize;
    s->samplerate   = samplerate;
    s->bps          = bps;
    s->decorrelation= decorrelation;

    return 0;
}

static u32 flac_decode_frame(struct FLACContext *s, s16 *pSamplesLRBuf)
{
    int i, j = 0;

    while(1){
      u8 b0;
      if(fread(&b0,1,1,FileHandle)!=1){
        printf("End of file.\n");
        return(0);
      }
      if(b0==0xff){
        u8 b1;
        if(fread(&b1,1,1,FileHandle)!=1){
          printf("End of file.\n");
          return(0);
        }
        if((b1&0xfe)==0xf8) break;
        fseek(FileHandle,ftell(FileHandle)-1,SEEK_SET);
      }
    }
    
    u32 toppos=ftell(FileHandle);
    
    uint8_t *bitstream= (u8*)malloc(s->max_framesize);
    u32 bitstreamsize=fread(bitstream,1,s->max_framesize,FileHandle);
//    printf("start frame: pos:$%x, data=$%02x,$%02x,$%02x,$%02x\n",toppos,bitstream[0],bitstream[1],bitstream[2],bitstream[3]);
    
    GetBitContext gb;
    init_get_bits(&gb, bitstream, bitstreamsize*8);

    if (decode_frame(s,&gb) < 0) {
        printf("FlacError: decode_frame() failed.\n");
        if(bitstream!=NULL){
          free(bitstream); bitstream=NULL;
        }
        fseek(FileHandle,toppos,SEEK_SET);
        return(0);
    }

    /* subframes */
    for (u32 i = 0; i < s->channels; i++) {
        if (decode_subframe(s, &gb, i) < 0){
          if(bitstream!=NULL){
            free(bitstream); bitstream=NULL;
          }
          return -1;
        }
    }

    align_get_bits(&gb);

    /* frame footer */
    skip_bits(&gb, 16); /* data crc */

#define DECORRELATE(left, right) {\
            assert(s->channels == 2);\
            if (s->is32) {\
              printf("%d: Can not 32bits.\n",__LINE__); \
              }else{\
              for (i = 0; i < s->blocksize; i++) {\
                int a= s->decoded[0][i];\
                int b= s->decoded[1][i];\
                *pSamplesLRBuf++ = (left)  << s->sample_shift;\
                *pSamplesLRBuf++ = (right) << s->sample_shift;\
              }\
            }\
          }

  if(pSamplesLRBuf!=NULL){
    switch (s->decorrelation) {
      case INDEPENDENT: {
        switch(s->channels){
          case 1: {
            if(s->is32){
              printf("%d: Can not 32bits.\n",__LINE__);
              }else{
              for (j = 0; j < s->blocksize; j++) {
                s16 smp=s->decoded[0][j] << s->sample_shift;
                *pSamplesLRBuf++ = smp;
                *pSamplesLRBuf++ = smp;
              }
            }
          } break;
          case 2: {
            if (s->is32){
              printf("%d: Can not 32bits.\n",__LINE__);
              }else{
              for (j = 0; j < s->blocksize; j++) {
                *pSamplesLRBuf++ = s->decoded[0][j] << s->sample_shift;
                *pSamplesLRBuf++ = s->decoded[1][j] << s->sample_shift;
              }
            }
          } break;
        }
      } break;
      case LEFT_SIDE: DECORRELATE(a,a-b); break;
      case RIGHT_SIDE: DECORRELATE(a+b,b); break;
      case MID_SIDE: DECORRELATE( (a-=b>>1) + b, a); break;
    }
  }

    u32 readedsize= (get_bits_count(&gb)+7)/8;
    
    if(bitstream!=NULL){
      free(bitstream); bitstream=NULL;
    }
  
    if (readedsize > bitstreamsize) {
        printf("FlacError: overread: %d\n", readedsize - bitstreamsize);
        return(0);
    }

    fseek(FileHandle,toppos+readedsize,SEEK_SET);
//    printf("%d, %d\n",readedsize,s->blocksize);
    
    return(s->blocksize);
}

// -----------------------------------------------------------------------------------------------------

u32 flac_decode(u32 *pSamplesLRBuf)
{
  u32 SamplesCount=flac_decode_frame(&s,(s16*)pSamplesLRBuf);
  if(SamplesCount==0) printf("End stream.\n");
  return(SamplesCount);
}

// -----------------------------------------------------------------------------------------------------

u32 flac_GetBlockSize(struct FLACContext *s, u8 *pbuf,u32 bufsize)
{
    int blocksize_code = *pbuf>>4;
    int sample_rate_code = *pbuf&0x0f;
    
    pbuf++;

    int assignment = *pbuf>>4; /* channel assignment */
    if (assignment < 8 && s->channels == assignment+1){
      }else{
      if (assignment >=8 && assignment < 11 && s->channels == 2){
        }else{
        printf("!");
        return -1;
      }
    }

    // skip get_bits(&gb, 3); sample_size_code
    
    if (*pbuf&0x01) {
        printf("!");
        return -1;
    }

    pbuf++;
    
    {
      s32 val=*pbuf++;
      int ones= 7 - ff_log2_tab[val ^ 255];
      if(ones==1){
        printf("!");
        return -1;
      }
      
      val&= 127>>ones;
      
      while(--ones > 0){
          int tmp= (*pbuf++) - 128;
          if(tmp>>6){
            printf("!");
            return -1;
          }
          val= (val<<6) + tmp;
      }
    }
    
    int blocksize;
    if (blocksize_code == 0)
        blocksize = s->min_blocksize;
    else if (blocksize_code == 6)
        blocksize = (*pbuf++)+1;
    else if (blocksize_code == 7){
        u32 val=(*pbuf++)<<8;
        val|=*pbuf++;
        blocksize = val+1;
    }else
        blocksize = blocksize_table[blocksize_code];

    if (blocksize > s->max_blocksize) {
        printf("!");
        return -1;
    }

    if (14<sample_rate_code){
        printf("!");
        return -1;
    }

    return(blocksize);
}

